import { getHighlighter } from './shiki/shiki';
import { cache, generateHash, codeLanguage, renderToHtml } from './utils';
import { selectAll } from 'unist-util-select';
import memoize from 'micro-memoize';
import pAll from 'p-all';

const DEFAULT_THEME = 'Material-Theme-Darker-High-Contrast';

const generateHighlighter = memoize(
  async () =>
    getHighlighter({
      theme: DEFAULT_THEME,
      renderer: renderToHtml,
      cache,
    }),
  { isPromise: true }
);

// @ts-ignore
const highlight = (options: any) => async (code: string, lang: string) => {
  const highlighter = await generateHighlighter();
  // @ts-ignore
  return highlighter.codeToHtml(code, lang as any);
};

const tryCachedNode = (options: any) => async (node: any) => {
  const hash = generateHash(node.value);
  const cached = await cache.get(hash);
  if (cached) {
    node.value = cached;
  } else {
    const lang = codeLanguage(node);
    node.lang = lang;
    const highlighted = await highlight(options)(node.value, lang);
    await cache.set(hash, highlighted);
    node.value = highlighted;
  }
};

const visitor = (options: any) => async (node: any) => {
  node.type = 'html';
  node.children = undefined;
  if (process.env.NODE_ENV === 'production') {
    // we cache nodes because they get highlighted multiple times during build
    // and we don't need to redo it each time
    await tryCachedNode(options)(node);
  } else {
    const lang = codeLanguage(node);
    node.lang = lang;
    node.value = await highlight(options)(node.value, lang);
  }
};

const transformer = (options: any) =>
  memoize(
    async (tree: any) => {
      const nodes = selectAll('code', tree);
      await pAll(
        nodes.map(node => () => visitor(options)(node)),
        { concurrency: 10 }
      );
      return tree;
    },
    { isPromise: true }
  );

function attacher(options: any) {
  return transformer(options);
}

module.exports = memoize(attacher);
